/*
 * find.c
 * Copyright (C) LenaD 2010 <lenad89@list.ru>
 * 
 */
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <malloc.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fnmatch.h>
#include <libgen.h>

void passgdir(char *ppath);
void mreaddir(DIR *dir, char *currpath);

char *name;
int homesize;

void *
xmalloc(size_t size)
{
	void *tmp = malloc(size);
	
	if (tmp == NULL) {		
		fprintf(stderr, "xmalloc: out of memory\n");
		exit(1);		
	}
	
	memset(tmp, 0, size);
	
	return tmp;	
}

char *
xstrdup(const char *str)
{
        char *rs;
        size_t len;

        len = strlen(str);

        rs = malloc(len+1);

        if (rs == NULL) {
                fprintf(stderr, "xstrdup: out of memory\n");
                exit(1);
        }

        memcpy(rs, str, len);
        rs[len] = 0;

        return rs;
}


char *
xgetcwd(void)
{
#define CWD_GROW_UP 32
        size_t len;
        char *rs, *s, *tmp;
	rs = NULL;
        len = 0;
        do {
                rs = realloc(rs, len+CWD_GROW_UP);
                if (rs == NULL) {
                        fprintf(stderr, "xgetcwd: out of memory\n");
                        exit(1);
                }
                len += CWD_GROW_UP;
                s = getcwd(rs, len);
        } while ((s == NULL) && (errno == ERANGE));

        if (s == NULL) {
                fprintf(stderr, "xgetcwd: can't get current dir: %s\n",
                                strerror(errno));
                exit(1);
        }
	
	return rs;
}

void correct(char *fname, int *iscorr)
{
	if (name != NULL) {
		char *bname = basename(fname);	
		int ismatch = fnmatch(name, bname, 0);
		if (ismatch == FNM_NOMATCH) {
			*iscorr = 0;
		} else if (ismatch != 0) {
			fprintf(stderr, "correct: wrong pattern");
			exit(1);
		}
	}
}

void push(char* fname)
{	
	if (homesize)
		printf(".%s\n", fname+homesize);
	else	
		printf("%s\n", fname);
}

void 
openderr(int sterr,char *fname)
{
	fprintf(stderr, "unable to open directory: "
	        "%s :%s\n",fname,strerror(sterr));
}
void staterror(int sterr,char *fname)
{
	fprintf(stderr, "could not stat :"
	        "%s :%s\n",fname,strerror(sterr));	 
}

void
passgdir(char *ppath)
{
	DIR* dir;
	struct stat pst;
	int isadd = 1;
	if (lstat(ppath, &pst) == -1) {
		staterror(errno, ppath);
	} else {		

		correct(ppath, &isadd);
		
		if (isadd)
			push(ppath);
			
		if (S_ISDIR(pst.st_mode)) {
			if ((dir = opendir(ppath)) != NULL) {
				mreaddir(dir, ppath);
				closedir(dir);
			} else
				openderr(errno, ppath);		
		}
	}
}

void
mreaddir(DIR *dir,char *currpath)
{
	struct dirent *currfile;
	char *fullname;
	size_t sizep = strlen(currpath);
	
	while ((currfile = readdir(dir)) != NULL) {
		
		if (strcoll(currfile->d_name, ".") && strcoll(currfile->d_name, "..")) {
			fullname = (char *)xmalloc(strlen(currfile->d_name)+sizep+2);
			sprintf(fullname, "%s/%s", currpath, currfile->d_name);
			passgdir(fullname);
			free(fullname);	
		}		
				
	}
}

void 
curpath(char *path)
{
	if (path[strlen(path)-1] == '/')
		path[strlen(path)-1] = 0;
	passgdir(path);
	free(path);
}
void
go_path(int cpath, char **argv)
{	
	int i;
	size_t len;
	char *tmp = xgetcwd();
	char *path = NULL;
	len = strlen(tmp);
	
	if (cpath > 0) {
		for(i = 0; i < cpath; i++){
			if (argv[i+1][0] == '/'){
				path = xstrdup(argv[i+1]);	
				curpath(path);
			} else {
				len = len + strlen(argv[i+1]) + 2;
				path= (char *)xmalloc(len);
				snprintf(path, len, "%s/%s", tmp, argv[i+1]);
				curpath(path);
			}
		}
	} else {
		path = tmp;
		curpath(path);
	}
}

void get_opt(int numopt, int argc, char **argv)
{
	int check_name = 0;
	
	if (numopt < argc && !strcoll(argv[numopt], "-name")) {
		if (++numopt < argc && argv[numopt][0] != '-' ){	
			name = xstrdup(argv[numopt]);
			numopt++;	
		}
		check_name = 1;
	}
	if ( numopt < argc || (check_name && name == NULL  )) {
		fprintf(stderr, "wrong: find [path][-name:pattern]\n");
		exit(1);
	}	
}

int
main(int argc, char **argv)
{	
	int i;
	int numopt;
	int cpath;
	
	char *path;

	for(i = 1, cpath = 0; i < argc && argv[i][0] != '-'; i++)
		cpath++;	
	numopt = i;	
	get_opt(numopt, argc, argv);
	go_path(cpath, argv);
	free(name);
	
	return 0;
};
